#include <stdio.h>
#include <err.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <sys/ptrace.h>
#include <sys/syscall.h>
#include <linux/elf.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

struct kgsl_command_object {
	uint64_t offset;
	uint64_t gpuaddr;
	uint64_t size;
	unsigned int flags;
	unsigned int id;
};

struct kgsl_gpu_command {
	uint64_t flags;
	uint64_t cmdlist;
	unsigned int cmdsize;
	unsigned int numcmds;
	uint64_t objlist;
	unsigned int objsize;
	unsigned int numobjs;
	uint64_t synclist;
	unsigned int syncsize;
	unsigned int numsyncs;
	unsigned int context_id;
	unsigned int timestamp;
};

static int re_task_id = -1;

static int vm_read(void *dst, unsigned long src, unsigned long len) {
  struct iovec lv = { .iov_base = dst, .iov_len = len };
  struct iovec rv = { .iov_base = (void*)src, .iov_len = len };
  return (syscall(__NR_process_vm_readv, re_task_id, &lv, 1, &rv, 1, 0) == len);
}
#define VM_READ(dst_obj, src_addr) vm_read(&(dst_obj), (src_addr), sizeof((dst_obj)))

#define CP_TYPE4_PKT	(4U << 28)
#define CP_TYPE7_PKT	(7U << 28)
static inline uint pm4_calc_odd_parity_bit(uint val)
{
	return (0x9669 >> (0xf & ((val) ^
	((val) >> 4) ^ ((val) >> 8) ^ ((val) >> 12) ^
	((val) >> 16) ^ ((val) >> 20) ^ ((val) >> 24) ^
	((val) >> 28)))) & 1;
}
#define cp_type4_base_index_one_reg_wr(pkt) (((pkt) >> 8) & 0x7FFFF)
#define type4_pkt_size(pkt) ((pkt) & 0x7F)
#define pkt_is_type4(pkt) \
	((((pkt) & 0xF0000000) == CP_TYPE4_PKT) && \
	 ((((pkt) >> 27) & 0x1) == \
	 pm4_calc_odd_parity_bit(cp_type4_base_index_one_reg_wr(pkt))) \
	 && ((((pkt) >> 7) & 0x1) == \
	 pm4_calc_odd_parity_bit(type4_pkt_size(pkt))))
#define pkt_is_type7(pkt) \
	((((pkt) & 0xF0000000) == CP_TYPE7_PKT) && \
	 (((pkt) & 0x0F000000) == 0) && \
	 ((((pkt) >> 23) & 0x1) == \
	 pm4_calc_odd_parity_bit(cp_type7_opcode(pkt))) \
	 && ((((pkt) >> 15) & 0x1) == \
	 pm4_calc_odd_parity_bit(type7_pkt_size(pkt))))
#define cp_type7_opcode(pkt) (((pkt) >> 16) & 0x7F)
#define type7_pkt_size(pkt) ((pkt) & 0x3FFF)
static void scan_insns(unsigned long base_addr, unsigned long insns_bytes) {
  uint32_t *insns = malloc(insns_bytes);
  if (!insns) err(1, "malloc");
  if (!vm_read(insns, base_addr, insns_bytes))
    goto out;
  unsigned long num_insns = insns_bytes / 4;
  unsigned long insn_idx = 0;
  while (insn_idx < num_insns) {
    uint32_t insn = insns[insn_idx];
    if (pkt_is_type7(insn)) {
      if (cp_type7_opcode(insn) == 0x3f/*CP_INDIRECT_BUFFER*/ && type7_pkt_size(insn) == 3) {
        uint32_t *p = insns + insn_idx + 1;
        uint64_t target_addr = p[0] | ((uint64_t)p[1] << 32);
        uint32_t target_size = p[2];
        scan_insns(target_addr, target_size*4);
      }
      insn_idx += 1 + type7_pkt_size(insn);
    } else if (pkt_is_type4(insn)) {
      uint32_t base_reg = cp_type4_base_index_one_reg_wr(insn);
      uint32_t num_regs = type4_pkt_size(insn);
      if (base_reg <= 0x88d8/*RB_BLIT_DST*/ && base_reg+num_regs > 0x88d8+1) {
        uint32_t *addrp = insns + insn_idx + 1 + (0x88d8 - base_reg);
        uint64_t addr = addrp[0] | ((uint64_t)addrp[1] << 32);
        printf("RB_BLIT_DST = 0x%lx\n", addr);
      }
      insn_idx += 1 + num_regs;
    } else {
      printf("bad insn type\n");
      goto out;
    }
    if (insn_idx > num_insns) {
      printf("bad insn length\n");
      goto out;
    }
  }
out:
  free(insns);
}

int main(void) {
  setbuf(stdout, NULL);

  printf("searching for surfaceflinger\n");
  DIR *proc = opendir("/proc");
  if (proc == NULL) errx(1, "open /proc");
  int surfaceflinger_pid;
  while (1) {
    struct dirent *dent = readdir(proc);
    if (!dent)
      errx(1, "unable to find surfaceflinger");
    if (dent->d_name[0] < '0' || dent->d_name[0] > '9')
      continue;
    int pid = atoi(dent->d_name);
    char comm_path[200];
    sprintf(comm_path, "/proc/%d/comm", pid);
    int comm_fd = open(comm_path, O_RDONLY);
    if (comm_fd == -1)
      continue;
    char comm_buf[100];
    int comm_len = read(comm_fd, comm_buf, sizeof(comm_buf)-1);
    close(comm_fd);
    if (comm_len <= 0)
      continue;
    comm_buf[comm_len] = '\0';
    if (strcmp(comm_buf, "surfaceflinger\n") != 0)
      continue;
    surfaceflinger_pid = pid;
    break;
  }
  closedir(proc);
  printf("surfaceflinger is pid %d\n", surfaceflinger_pid);

  char tasks_path[200];
  sprintf(tasks_path, "/proc/%d/task", surfaceflinger_pid);
  DIR *tasks = opendir(tasks_path);
  if (tasks == NULL) errx(1, "open tasks");
  while (1) {
    struct dirent *dent = readdir(tasks);
    if (!dent) errx(1, "unable to find thread");
    if (dent->d_name[0] < '0' || dent->d_name[0] > '9')
      continue;
    int tid = atoi(dent->d_name);
    char comm_path[200];
    sprintf(comm_path, "/proc/%d/task/%d/comm", surfaceflinger_pid, tid);
    int comm_fd = open(comm_path, O_RDONLY);
    if (comm_fd == -1)
      continue;
    char comm_buf[100];
    int comm_len = read(comm_fd, comm_buf, sizeof(comm_buf)-1);
    close(comm_fd);
    if (comm_len <= 0)
      continue;
    comm_buf[comm_len] = '\0';
    if (strcmp(comm_buf, "RenderEngine\n") != 0)
      continue;
    re_task_id = tid;
    break;
  }
  printf("RenderEngine is tid %d\n", re_task_id);

  if (ptrace(PTRACE_ATTACH, re_task_id, NULL, NULL)) err(1, "ptrace attach");
  int status;
  if (waitpid(re_task_id, &status, 0) != re_task_id) err(1, "wait for attach");

  while (1) {
    if (ptrace(PTRACE_SYSCALL, re_task_id, NULL, NULL)) err(1, "ptrace syscall entry");
    if (waitpid(re_task_id, &status, 0) != re_task_id) err(1, "wait for entry");

    struct user_pt_regs syscall_regs;
    struct iovec sysc_regs_vec = { .iov_base = &syscall_regs, .iov_len = sizeof(syscall_regs) };
    if (ptrace(PTRACE_GETREGSET, re_task_id, (void*)NT_PRSTATUS, &sysc_regs_vec)) err(1, "ptrace get regs");
    if (syscall_regs.regs[8] == __NR_ioctl && (syscall_regs.regs[1] & 0xffff) == 0x094A) {
      unsigned long kgsl_gpu_command_addr = syscall_regs.regs[2];
      struct kgsl_gpu_command kgc;
      if (!VM_READ(kgc, syscall_regs.regs[2]))
        goto next;

      for (int i=0; i<kgc.numcmds; i++) {
        struct kgsl_command_object kco;
        if (!VM_READ(kco, kgc.cmdlist + i * kgc.cmdsize))
          goto next;
        scan_insns(kco.gpuaddr, kco.size);
      }
    }

next:
    if (ptrace(PTRACE_SYSCALL, re_task_id, NULL, NULL)) err(1, "ptrace syscall exit");
    if (waitpid(re_task_id, &status, 0) != re_task_id) err(1, "wait for exit");
  }
}
